module PUBMED where

import Conduit
import Control.Applicative
import Control.Concurrent (threadDelay)
import Data.Attoparsec.ByteString
import Data.Attoparsec.ByteString.Char8 (anyChar)
import Data.ByteString.Char8 (pack)
import Data.ByteString.Lazy (ByteString)
import Data.Maybe (fromMaybe)
import Data.Text (Text)
import Network.HTTP.Client (newManager)
import Network.HTTP.Client.TLS (tlsManagerSettings)
import PUBMED.Client
import PUBMED.Parser
import Panic (panic)
import Prelude hiding (takeWhile)
import Servant.Client (runClientM, mkClientEnv, BaseUrl(..), ClientEnv, ClientError, Scheme(..))
import qualified Data.ByteString.Lazy as LBS
import qualified Data.List as List
import qualified Data.Text as T
import qualified Data.Text.Encoding as TE
import qualified Data.Text.Lazy as TL

pmHost :: String
pmHost = "eutils.ncbi.nlm.nih.gov"
pmSearchPath :: String
pmSearchPath = "entrez/eutils"
pmPort :: Int
pmPort = 443

batchSize :: Int
batchSize = 200

defaultEnv :: IO ClientEnv
defaultEnv = do
  manager' <- newManager tlsManagerSettings
  pure $ mkClientEnv manager' $ BaseUrl Https pmHost pmPort pmSearchPath

-- | API main function
getMetadataWith :: Maybe Text -> Text -> Maybe Integer -> Maybe Limit -> IO (Either Text [PubMed])
getMetadataWith = runSimpleFindPubmedAbstractRequest

getMetadataWithC :: Maybe Text -> Text -> Maybe Limit -> IO (Either ClientError (Maybe Integer, ConduitT () PubMed IO ()))
getMetadataWithC mAPIKey query limit = do
  env <- defaultEnv
  -- First, estimate the total number of documents
  eRes <- runClientM (search mAPIKey (Just query) Nothing (Just 1)) env
  pure $ get' env query limit batchSize <$> eRes
  where
    get' :: ClientEnv -> Text -> Maybe Limit -> Int -> BsXml -> (Maybe Integer, ConduitT () PubMed IO ())
    get' env q l perPage (BsXml res) =
      ( Just numResults
      , yieldMany [0..]
        .| takeC (fromInteger numPages)
        .| concatMapMC (getPage env mAPIKey q perPage))
      where
        numResults = fromMaybe 0 $ parseDocCount $ TL.fromStrict $ TE.decodeUtf8 $ LBS.toStrict res
        numPages = numResults `div` (fromIntegral perPage) + 1

getPage :: ClientEnv
        -> Maybe Text
        -> Text
        -> Int
        -> Int
        -> IO [PubMed]
getPage env mAPIKey q perPage pageNum = do
  let offset = fromIntegral $ pageNum * perPage
  eDocs <- runSimpleFindPubmedAbstractRequest mAPIKey q (Just offset) (Just $ fromIntegral perPage)
  case eDocs of
    Left err -> panic $ "[getPage] error: " <> show err
    Right docs -> do
      _ <- threadDelay 2000000 -- One seconds
      pure docs

-- | TODO this parser need at least one subs at the end
-- (use endOfInput)
removeSub :: Parser ByteString
removeSub = do
  dt <- many textWithBalise
  eo <- manyTill anyChar endOfInput
  pure $ LBS.fromStrict $ pack $ concat dt <> eo
  where
    textWithBalise =
      manyTill anyChar (sub <|> asub)
    sub  = string "<sub>"
      <|> string "<sup>"
      <|> string "<i>"
      <|> string "<b>"
    asub = string "</sub>"
      <|> string "</sup>"
      <|> string "</i>"
      <|> string "</b>"

type Query = Text
type Limit = Integer

runMultipleFPAR :: [Integer]
                -> IO (Either Text [PubMed])
runMultipleFPAR [] = pure $ Right []
runMultipleFPAR ids = List.foldl1' concat'
                   <$> mapM runSimpleFetchPubmedAbstractRequest (by 300 ids)
  where
    by _ [] = []
    by n ns = head' : ( by n tail')
      where
        (head',tail') = List.splitAt n ns

    concat' (Right n) (Right m) = Right $ n <> m
    concat' n m = n <> m

runSimpleFetchPubmedAbstractRequest ::
  [Integer]
  -> IO (Either Text [PubMed])
runSimpleFetchPubmedAbstractRequest ids = do
  env <- defaultEnv
  res <- runClientM
    (fetch Nothing (Just "pubmed") (Just "abstract") ids)
    env
  case res of
    (Left err) -> pure (Left . T.pack $ show err)
    (Right (BsXml abs)) -> do
      --putStrLn $ show abs
      case parseOnly removeSub $ LBS.toStrict abs of
        (Left err'') -> pure (Left $ T.pack err'')
        (Right v) -> do
          let parsed = parsePubMed $ TL.fromStrict $ TE.decodeUtf8 $ LBS.toStrict v
          pure $ Right parsed
--          parsed <- catch (pubMedParser v) ((\e -> do
--                                                _ <- print e
--                                                pure []) :: XmlException -> IO [PubMed])
--          Right <$> pure parsed

runSimpleFindPubmedAbstractRequest :: Maybe Text
                                   -> Text
                                   -> Maybe Integer
                                   -> Maybe Limit
                                   -> IO (Either Text [PubMed])
runSimpleFindPubmedAbstractRequest mAPIKey query offset limit =  do
  eDocIds <- searchDocIds mAPIKey query offset limit
  case eDocIds of
    Left err -> pure $ Left err
    Right docIds -> runMultipleFPAR docIds

searchDocIds :: Maybe Text -> Text -> Maybe Integer -> Maybe Limit -> IO (Either Text [Integer])
searchDocIds mAPIKey query offset limit = do
  env <- defaultEnv
  res <- runClientM
    (search mAPIKey (Just query) offset limit)
    env
  case res of
    (Left err) -> pure (Left $ T.pack $ show err)
    (Right (BsXml docs)) -> do
      pure $ Right $ parseDocIds $ TL.fromStrict $ TE.decodeUtf8 $ LBS.toStrict docs
      --pure $ Right $ runParser parseDocId docs

--runParser :: Show res => (Cursor -> res) -> LBS.ByteString -> res
--runParser parser = parser . fromDocument . parseLBS_ def
